/*
*	name: formparse
*	function: decode a form-data and store the result in the file
*	usage:
*		formparse [-Er] [-o dir] [query] 
*		formparse -r -o dir query   # form-data is given by query
*		or
*		formparse -r -o dir	# form-data is given by stdin
*		---
*		option -E: simplly echo message body for debug
*		option -r: removes CR codes from the data
*
*	result:
*		decoded data is stored in files. the direcory structure is:
*		name/N	where N is 0, 1, ...
*		the count of duplicated name is in
*		name/clone
*
*	example1: in case of get-method
*		{ramfs >[2=1]} >/dev/null
*		formparse -r -o /tmp $QUERY_STRING
*
*	example2: in case of post-method
*		{ramfs >[2=1]} >/dev/null
*		formparse -r -o /tmp
*
*	note:
*		you will find some codes for multipart/form-data,
*		however they are still under construction.
*
*	auther: Kenji Arisawa <arisawa@aichi-u.ac.jp>
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>

#define NAMELEN 128

enum{
	EOD	= -1,
	LF = 0x0a,
	CR = 0x0d,
};

typedef struct namval namval;
struct namval{
	char *name;
	char *value;
};

static void error(char *fmt,...);
static int x2i(int i);
static char *qparse(char *qs);
static int cparse(char *b);
static void mparse(char *b);
static void xparse(char *dir);
static char *skipto(char *s, char *p);
static char *trim(char *s, char *p);
static char *value(char *s, char *name);
static Biobuf *Bnew1(char *name);
static Biobuf *Bnew2(char *dir, char *name);
static int readto(int fd, char *bp, int size, char del);
static int	getmime(char *t, namval *ps, int max);

static int writefile(char *fn, char *b, int n);
static int readfile(char *fn, char *b, int n);
shapeup= 0;

char *odir="/tmp";

static void
usage(void)
{
	fprint(2,"usage: formparse [ -r] [-o dir] [query]\n");
	exits("usage");
}

static void
error(char *fmt,...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, fmt);
	vseprint(buf, buf+sizeof(buf), fmt, arg);
	va_end(arg);
	print("HTTP/1.0\nContent-type: text/html\n\n");
	print("<html><head><title>Form processing error</title></head>");
	print("<body><h1>Error</h1><b>%s</b></body></html>\n", buf);
	exits(buf);
}

void
main(int argc, char *argv[])
{
	char *qs;
	char *s, *t, *p;
	char b[80];
	int echoflag = 0;
	int n;

	ARGBEGIN{
	case 'o':
		odir=ARGF();
		break;
	case 'r':
		shapeup= 1;
		break;
	case 'E': // simply echo message body for debug
		echoflag=1;
		break;
	default:
		usage();
	}ARGEND
	if(chdir(odir) < 0)
		error("could not chdir to %s",odir);
	qs = nil;
	if(argv[0])
		qs = argv[0];
	if(qs)// parse x-www-form-urlencoded string
		while((qs = qparse(qs)));
	else{
		 s = getenv("HTTP_CONTENT_LENGTH");
		if(s == nil)
			error("no content length");
		n = atoi(s);
		if(echoflag){
			print("content-length=%d\n", n);
			s = malloc(n+1);
			n = read(0, s, n);
			print("read-length=%d\n", n);
			write(1, s, n);
			exits(nil);
		}
		t = getenv("HTTP_CONTENT_TYPE");
		if(t == nil)
			error("no content type");
		p = skipto(t,";");
		if(*p == ';')
			*p++ = 0;
		t = trim(t, " ");
		if(strcmp(t,"application/x-www-form-urlencoded") == 0){
			qs = malloc(n + 1);
			n = readn(0, qs, n);
			qs[n] = 0;
			while((qs = qparse(qs)));
		}
		else if(strcmp(t,"multipart/form-data") == 0){
			p = value(p, "boundary");
			snprint(b, sizeof b, "%c--%s", LF, p); // see the note on cparse for LF
			mparse(b);
		}
		else
			error("unknown format: %s", t);
	}
}

/*	conver hex char to int	*/
static int
x2i(int i)
{
	i=toupper(i);
	i = i - '0';
	if(i > 9) i = i - 'A' + '9' + 1;
	return i;
}

static int
readfile(char *fn, char *b, int n)
{
	int fd;
	fd = open(fn, OREAD);
	if(fd < 0)
		return -1;
	n = read(fd, b, n);
	b[n] = 0;
	close(fd);
	return n;
}

static int
writefile(char *fn, char *b, int n)
{
	int fd;
	fd = open(fn, OWRITE);
	if(fd < 0)
		return -1;
	n = write(fd, b, n);
	close(fd);
	return n;
}

static int
mkdir(char *name)
{
	Dir *dir;
	int fd;
	if((dir = dirstat(name)) == nil){ /* the name is first apeared */
		fd = create(name, OREAD, 0750 | DMDIR);
		if(fd < 0) /* somthing wrong */
			error("could not create %s", name);
		close(fd);
		return 0;
	}
	free(dir);
	return -1;
}


static Biobuf *
Bnew1(char *name)
{
	Biobuf *bout;
	char buf[20], *fn;
	int fd, n, fnsize;
	/*
	*	we must count the occurrence of the name.
	*	count it in:
	*		name/clone
	*/
	fnsize = strlen(name)+10;
	fn = malloc(fnsize);
	if(access(name,0) == -1)
		mkdir(name);// how we should handle the status?
	snprint(fn, fnsize,"%s/clone",name);

	if(access(fn,0) == -1){
		fd = create(fn, OWRITE, 0640);
		if(fd < 0) /* somthing wrong */
			error("%s not open", fn);
		write(fd,"0",1);
		close(fd);
	}
	/*
	*	read clone. increase the value. write it back
	*	the name and the value is sored in:
	*	name/0, name/1, ...
	*/
	if(readfile(fn, buf, sizeof(buf))<0)
		error("cannot read %s", fn);

	snprint(fn, fnsize,"%s/%s", name, buf);
	bout = Bopen(fn, OWRITE);
	if(!bout)
		error("%s not open", fn);
	n = atoi(buf) + 1;
	snprint(buf,sizeof(buf),"%d", n);
	snprint(fn, fnsize,"%s/clone",name);
	if(writefile(fn, buf, strlen(buf)) < 0)
		error("cannot write to %s", fn);
	free(fn);
	return bout;
}

static Biobuf *
Bnew2(char *dir, char *name)
{
	Biobuf *b;
	char *fn;
	int fnsize;
	mkdir(dir);
	fnsize = strlen(dir) + strlen(name) + 10;
	fn = malloc(fnsize);
	snprint(fn, fnsize,"%s/%s",dir, name);
	b = Bopen(fn, OWRITE);
	if(b == nil) /* somthing wrong */
		error("%s not open", fn);
	free(fn);
	return b;
}

/*
*	parse single name and the value
*	return:
*	0:	End of data or fatal error
*	else:	next point to parse; parsing single name/value pair has finished
*/
static char *
qparse(char *qs)
{
	char *s, *s0;
	char name[NAMELEN];
	int c, c1;
	Biobuf *bout;

	/*
	*	regular format is,
	*	name=val&name=val&...name=val
	*	but the data can be forged, so we must be careful.
	*	example of forged patterns:
	*	&&&&
	*	====
	*	=&=&=
	*/

	/*
	*	parse a name
	*	we restrict chars in the name to be alnum
	*/
	s = s0 = name;
	while((c = *qs++)){
		if(s - s0 == NAMELEN - 1){ /* truncate here */
			break;
		}
		if(c == '='){
			break;
		}
		if(isalnum(c)){
			*s++ = c;
			continue;
		}
		/* something wrong. skip it */
	}
	*s = 0;
	if(c == 0)
		return 0;

	/* sweep out extra chars */
	while(c != '=' && (c = *qs++));

	/* check the name length */
	if(strlen(name) == 0){ /*something wrong. skip to next name */
		while((c = *qs++) && c != '&');
		return qs;
	}

	bout = Bnew1(name);

	/* get the value */
	while((c = *qs++)){
		switch(c){
		case '&':
			Bterm(bout);
			return qs;
		case '+':
			c = ' ';
			break;
		case '%':
			c = *qs++;
			if(c == 0) goto err;
			c1 = c;
			c = *qs++;
			if(c == 0) goto err;
			c = 16*x2i(c1) + x2i(c);
			if(c == '\r' && shapeup) continue;
			break;
		default:
			;
		}
		Bputc(bout, c);
	}
err:
	Bterm(bout);
	return 0;
}


/*
	Ref: "Form-based File Upload in HTML" (RFC1867)
	Example

<form enctype="multipart/form-data" method="POST" action="getform3.html"><br>
Your Name
<input type="text" name="name" value="alice" size="10"><br>
<input type="text" name="name" value="bod" size="10"><br>
<input type="text" name="name" value="carol" size="10"><br>
<input type="file" name="file" size="60"><br>
<input type="reset" value="RESET"><input type="submit" value="OK">
</form>




content-length=1259
read-length=1259
--#140133247#multipart#boundary#1149474368#
Content-Disposition: form-data; name="name"

alice
--#140133247#multipart#boundary#1149474368#
Content-Disposition: form-data; name="name"

bod
--#140133247#multipart#boundary#1149474368#
Content-Disposition: form-data; name="name"

carol
--#140133247#multipart#boundary#1149474368#
Content-Disposition: form-data; name="file"; filename=""; :"
Content-Type: www/unknown

[file body here]

--#140133247#multipart#boundary#1149474368#--
*/

/*
        Content-disposition: attachment; filename="file2.gif"
        Content-type: image/gif
        Content-Transfer-Encoding: binary
*/


/*
*	Note
*	must work for filename="a_file ;";": in it"
*	If it can work, the filename must be last in the line
*/


/*
*	parse multipart/form-data
*	currentlly only single file transmission is supported
*/
static void
mparse(char *b)
{
	char buf[256];
	int n,m,bsize;
	m = sizeof buf;
	bsize = strlen(b);

	buf[0] = LF;
	n = readto(0,buf+1, m-1, CR) + 1;
	if(buf[n-1] != CR)
		error("first boundary too long: len=%d", n);
	buf[n-1] = 0;
	if(strncmp(buf, b, bsize) != 0)
		error("not start with boundary");
	while(cparse(b) > 0);
}

/*
*	parse a single body of multipart/form-data
*	b: boundary string that start with LF
*
*	The format is:
*		content-disposition: form-data; name="..." CRLF
*		CRLF
*		... CRLF
*		boundary CRLF
*		content-disposition: form-data; name="..." file="..." CRLF
*		CRLF
*		... CRLF
*		boundary-- CRLF
*
*	where ... are some data and boundary is:
*	--boundary_string
*
*	Be aware that binary data may contain CRLF
*	We cannot undo write operation to file.
*	Therefore be carefull not write CRLF followed by boundary
*
*	To simplify processing, we define nboundary:
*		sprint(nboundary, sizeof nboundary,"%c%s", LF, boundary);
*	then we can read data using readto(0,buf,sizeof buf, CR)
*	argument b of cparse is nboundary
*/
static  int
cparse(char *b)
{
	char buf[4096], cr[1];
	namval ps[3];
	char *s;
	int n,m,bsize, f;
	Biobuf *bout;
	cr[0] = CR;
	bsize = strlen(b);
	m = sizeof buf;
	n = readto(0,buf, m, CR);
	if(buf[n-1] != CR)
		error("first line too long in content block");
	buf[n-1] = 0;
	s = trim(buf," \t\n");
	s = strdup(s);
	n = getmime(s, ps,3);
	if(cistrcmp(ps[0].name, "Content-Disposition") != 0)
		error("error: [%s] should be Content-Disposition", ps[0].name);
	if(strcmp(ps[0].value, "form-data") != 0)
		error("error: [%s] should be `form-data'", ps[0].value);
	if(n < 2)
		error("error: too few fields [%s]", buf);
	if(strcmp(ps[1].name,"name") != 0)
		error("error: [%s] should be `name'", ps[1].name);
	bout = nil;
	switch(n){
	case 2:
		bout = Bnew1(ps[1].value);
		break;
	case 3:
		if(strcmp(ps[2].name,"filename") != 0)
			error("error: [%s] should be `filename'", ps[2].name);
		/*	we must do something for some stupid browser 	*/
		/*	Netscape is wise enough, so we may not need but... */
		s = ps[2].value;
		s = strrchr(s, '/');
		if(s)
			ps[2].value = ++s;
		/*	convert ' ' -> '_'	*/
		s = ps[2].value;
		for(; *s; s++)
			if(*s == ' ')
				*s = '_';
		
		/*	stupid IE !	It pass me full path!
		*	we can guess Windows file from "[a-z][A-Z]:\*" */
		s = ps[2].value;
		if(strlen(s) > 2 && isalpha(s[0]) && s[1] == ':' && s[2] == '\\'){
			s = strrchr(s, '\\');
			if(s)
				ps[2].value = ++s;
		}
		
		/*	Windows pass me in SJIS code. How should I do?	*/

		bout = Bnew2(ps[1].value, ps[2].value);
		break;
	default:
		error("too many fileds");
	}
	/* 	there may some other headers such as:
	*	Content-Type: www/unknown
	*	we may discard these headers
	*	exit if read error		*/
	do n = readto(0,buf, m, CR);
	while(n > 0 && (n != 2 || buf[0] != LF || buf[1] != CR));
	if(n <= 0)
		error("no empty line after Content-Disposition of %s", b);

	/* 	Note that there is a small possiblity that the CR is a part of
	*	form controll, that is, the CR might be followed
	*	by nboundary	*/
	readto(0,buf,1,LF); // discard first LF, that is a part of form controll
	f = 0; // flag that means CR is held
	while((n = readto(0, buf, m, CR)) > 0){
		if(strncmp(buf, b, bsize) == 0)
			break;
		if(f){	// flush held CR
			Bwrite(bout, cr,1);
			f = 0;
		}
		if(buf[n-1] == CR){
			Bwrite(bout, buf, n-1); // last CR must be held
			f = 1;
		}
		else // buffer full, no CR in it.
			Bwrite(bout, buf, n);
	}
	Bterm(bout);


	if(strncmp(buf + bsize, "--\r", 3) == 0)
		return 0; // finished
	if(strncmp(buf + bsize, "\r", 1) == 0)
		return 1; // continue
	return -1; // error
}


static char *
skip(char *s, char *p)
{
	while(*s && strchr(p, *s))
		s++;
	return s;
}

static char *
skipto(char *s, char *p)
{
	while(*s){
		if(strchr(p, *s))
			break;
		s++;
	}
	return s;
}

static char*
trim(char *s, char *p)
{
	char *t;
	while(strchr(p,*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(strchr(p,*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

static char *
value(char *s, char *name)
{
	char *t;
	t = skipto(s, "=");
	if(*t == 0)
		error("syntax error");
	*t++ = 0;
	s = trim(s, " ");
	if(strcmp(s, name) != 0)
		error("error in name %s", s);
	return trim(t, " \"");
}


/*
*	readto
*	usage: n = readto(fd, buf, bufsize, DEL)
*	Readto is just like read, but read chars until DEL (delimiter)
*	and returns the number of chars read.
*	If DEL is read in buf then:
*		n <= bufsize
*	and 
*		buf[n-1] == DEL.
*	If a line is larger than bufsize then:
*		n == bufsize
*	and
*		buf[n-1] != DEL.
*
*	-kenar-
*/
static int
readto(int fd, char *bp, int size, char del)
{
	static char buf[4*4096];
	static char *s;
	static int n=0;
	int k=0;
	int i;
	USED(bp);
	USED(size);
	for(;;){
		if(n == 0){
			n = read(fd, buf, sizeof(buf));
			s = buf;
			if(n <= 0){
				n = 0;
				if(k == 0) return -1;
				else return k;
			}
		}
		for (i=0; i < n &&  i < size && *s != del;  i++) *bp++ = *s++;
		n -= i;
		size -= i;
		k += i;
		/*
		*	here we have three cases:
		*	1.  size == 0		-> simply return
		*	2.  n == 0			-> reread
		*	3.  *s == '\n'		-> *bp++ = *s++
		*	keep this order in evaluation.
		*/
		if(size==0) break;
		if(n == 0) continue;
		/* then *s must be del and also n>0 && size>0 */
		*bp++ = *s++; n--; size--; k++;
		break;
	}
	return k;
}

/*
*	t is scrached
*/
static int
getmime(char *t, namval *ps, int max)
{
	char *p, *e;
	int n;
	int stop;
	if(max == 0)
		return 0;
	stop = 0;
	p = skipto(t,":");
	if(*p == 0)
		return -1;
	*p = 0;
	ps[0].name = trim(t," ");
	t = p + 1;
	p = skipto(t,";");
	if(*p == 0)
		stop = 1;
	*p = 0;
	ps[0].value = trim(t," ");
	if(stop || max == 1)
		return 1;
	t = p + 1;
	n = 1;
	e = t + strlen(t);
	while(t < e){
		p = skipto(t,"=");
		if(*p == 0)
			break;
		*p = 0;
		ps[n].name = trim(t, " ");
		t = p + 1;
		p = skipto(t,";");
		if(*p == 0)
			stop = 1;
		*p = 0;
		/* then trim(t," ") is quoted string such as "alice" */
		t = trim(t, " ");
		/* *t should be '\"' and *(t + strlen(t) - 1) should be '\"' */
		if(*t != '\"' || *(t + strlen(t) - 1) != '\"')
			return -1; // format error
		*(t + strlen(t) - 1) = 0;
		ps[n].value = t+1;
		n++;
		if(stop || n == max)
			break;
		t = p + 1;
	}
	return n;
}

	