#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <bio.h>
#include <ndb.h>

#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

#define setenv(var, value) if(value) putenv(var, value)

typedef struct Handle Handle;
struct Handle
{
	char *mimetype;
	char **argv;
	int	ramfs;
};

typedef struct ChunkBuf ChunkBuf;
struct ChunkBuf
{
	uchar	buf[Hsize+32];
	int	size;
	ChunkBuf *next;
};

/*	basic constant	*/
const char *serversoft="Pegasus/2.0 (Plan9)";
const char *HTTPLOG = "http";
static const char *httpdconf = "/sys/lib/httpd.conf";
static const char *httpport="80";
static const char *httpsport="443";
static const char *netdir = "/net";

/*	default options	*/
static char *webbase = "/usr/web";
static char *namespace = "/lib/namespace.httpd";
static char *rewrite = "/sys/lib/httpd.rewrite"; // system rewrite-file
char *port = "80"; // replaced later by really accepted port. used in log.c
static int	server = 0; // 0: runs under the control of listen. 1: runs in server mode.
static int mntok=0;
static char *runas="none";
static int usermode = 0; // 1: httpd service by the user who invoked httpd
static int monmode = 0; // 1: not run as daemon; monitored by `mon'
static int noweb = 0; // 0: allows $home/web as HomePage if present

/*	the bellow may be required tuning for your server
*	and can be reset in configuration file "/sys/lib/httpd.conf"	*/
static int parsetimeout = 15; // timeout to parse header (in unit of sec)
static int alivetimeout = 15; // keep alive timeout (in unit of sec)
static int cgitimeout = 5; // timeout for CGI (in unit of sec)
static int posttimeout = 15*60; // timeout for post data (in unit of sec)
static int maxpost = 10; // in unit of MB. post data size limit that must less than 512MB(comes from ramfs limit)
static int connectlimit = 300; // limit of total open tcp connection
static int maxconnect = 20; // max tcp connection by the IP
static int obstime = 3; // observation time to detect burst access (in unit of sec)
static int maxaccess = 20; // allowed max access in obstime
static int contmax = 100; // max persistent continuation count for safty

/*	basic constant that are given later	*/
static char *myip; //  server IP. 202.250.160.40 for example
static char *myname; // server name. example: ar in ar.aichi-u.ac.jp
static char *padom; // parent dom. example: aichi-u.ac.jp
static char *mydom; //	myname + "." + padom
static char *scheme; // http or httpd
char *uriconnection; // "http://ar.aichi-u.ac.jp/" ...  # scheme://host

/* used internally for control */
static int cpid; // pid of CGI
static int mnt; // namespace configuration is finished
static int cont; // current persistent connection continuation count
static int notestatus; // alarm note to CGI is received
static char *webmnt; // to be "/usr/web/mnt"
static int binding; // 0: main document, 1: virtual host doc, 2: user's doc
static char *certificate;
static uchar *cert;
static int certlen;
static int ntossed; // current tossed count that controls redirection loop in persitent connection.

/*	Note on cont
*	cont is used as a flag so that we can detect re-request using Keep-Alive
*	if (cont > 0)
*		re-request due to Kepp-Alive
*		so we should not reconfigure our namespace
*		"cont" means "continued"
*/

static	char *csquery(char*, char*, char*);
static	char *sysdom(void);
static	char *pardom(void);
static	void		becomenone(void);
static	int		nsconfig(HConnect *c, char *webbase);
static	void		dolisten(char*);
static	int		doreq(HConnect*);
static	int		send(HConnect*);
static 	void		doservice(char *ndir);
static	int		acceptpost(HConnect *c, int timeout);
static 	int		execute(HConnect *c, char **argv, char *mimetype);
static	int		hparse2env(HConnect *c, char *head);
static 	int		notefun(void *a, char *msg);
static 	int	execsafely(HConnect *c, char *path, char **argv, char *mimetype);
static	int		merge(char *mnt, char *root);
static	int	ipcheck(HConnect *c, char *file, char *ip, char *path);
static 	char*	redirect(HConnect *c, char *file, char *path, char **webroot);
static 	Handle*	handler(HConnect *c, char *file, char *path);
static	void		setcgienv(HConnect *c);
static 	char *	hcat2(HConnect *c, char *a1, char *a2);
static	char *	hcat3(HConnect *c, char *a1, char *a2, char *a3);
static	int		ramfs(HConnect *c);
static 	int		nomount(HConnect *c);
static	char *	addoptions(HConnect *c, char *uri);
static	void		printheaders(HConnect *c, char *status, char *mimetype, int length);
int	creject(int ctl, char *ndir);

char *usage=
	"usage: httpd [-Mmsuw] [-t timeout] [-p port] [-c certificate] [-n namespace] [-R rewrite]  [[httpd] dummy ..  incalldir]\n";

void
main(int argc, char **argv)
{
	char buf[256];
	char *incalldir = nil;
	fmtinstall('D', hdatefmt);
	fmtinstall('H', httpfmt);
	fmtinstall('U', hurlfmt);
	if(readconfig(httpdconf) < 0){
		syslog(0, HTTPLOG, "# error in reading config # %r");
		exits("readconfig");
	}
	setstrconfig(&myname,"myname");
	setstrconfig(&webbase,"base");
	setstrconfig(&namespace,"namespace");
	setstrconfig(&rewrite,"rewrite");
	setintconfig(&alivetimeout, "alivetimeout");
	setintconfig(&parsetimeout, "parsetimeout");
	setintconfig(&cgitimeout, "cgitimeout");
	setintconfig(&posttimeout, "posttimeout");
	setintconfig(&maxpost, "maxpost");
	maxpost *= 1024*1024;
	setintconfig(&connectlimit, "connectlimit");
	setintconfig(&maxconnect, "maxconnect");
	setintconfig(&obstime, "obstime");
	setintconfig(&maxaccess,"maxaccess");
	setintconfig(&contmax,"contmax");
	ARGBEGIN{
	case 'n':
		namespace = ARGF();
		break;
	case 'p':
		port = ARGF(); /* http */
		break;
	case 'c':
		certificate = ARGF();
		cert = readcert(certificate, &certlen);
		if(cert == nil)
			sysfatal("reading certificat: %r");
		break;
	case 't':
		cgitimeout = atoi(ARGF());
		break;
	case 'R':
		rewrite = ARGF();
		break;
	case 's':
		server = 1;
		break;
	case 'm':
		mntok = 1;
		break;
	case 'u':
		usermode = 1;
		break;
	case 'w':
		noweb = 1;
		break;
	case 'M':
		monmode = 1;
		break;
	default:
		sysfatal(usage);
		break;
	}ARGEND

	if(argc > 0)
		incalldir = argv[argc-1];
	if((server && incalldir) || (!server && !incalldir))
		sysfatal("# error: inconsistent use of -s and incalldir: %s", usage);
	if(!server && monmode)
		sysfatal("# error: inconsistent use of -M: %s", usage);

	if(webbase[0] != '/')
		sysfatal("webbase must begin with '/'");
	
	snprint(buf, sizeof buf, "%s/mnt", webbase);
	webmnt = estrdup(buf);

	if(namespace[0] != '/')
		sysfatal("namespace must begin with `/'");

	if(rewrite[0] != '/')
		sysfatal("rewrite must begin with `/'");

	padom = pardom(); // parent dom; aichi-u.ac.jp in ar.aichi-u.ac.jp
	if(padom == nil) // something wrong
		sysfatal("pardom() returned nil");
	if(myname == nil)
		myname = sysname();
	snprint(buf, sizeof buf, "%s.%s", myname, padom);
	mydom = estrdup(buf);
	scheme = cert?"https":"http";
	if(!server){
		if(usermode)
			sysfatal("inconsistent options");
		time(nil);
		contentinit();
		urlinit();
		statsinit();
		doservice(incalldir);
		exits(nil);
	}

	if(!monmode){
		switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
		case -1:
			sysfatal("fork");
		case 0:
			break;
		default:
			exits(nil);
		}
	}

	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, HTTPLOG, nil);

	contentinit();
	urlinit();
	statsinit();

	runas = getuser();
	if(!usermode && (strcmp(runas, "none") != 0)){
		runas = "none";
		if(!certificate)
			becomenone();
	}
	dolisten(netmkaddr("*", "tcp", port));
	exits(nil);
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}

static HConnect*
mkconnect(void)
{
	HConnect *c;

	c = ezalloc(sizeof(HConnect));
	c->hpos = c->header;
	c->hstop = c->header;
	c->replog = 0;
	return c;
}

static HSPriv*
mkhspriv(void)
{
	HSPriv *p;

	p = ezalloc(sizeof(HSPriv));
	return p;
}

static void
dolisten(char *address)
{
	char ndir[NETPATHLEN], dir[NETPATHLEN], *p;
	int n, m, k, t, t0, ctl, nctl, data, spotchk;
	TLSconn conn;
	char		netdirb[256];
	char *ip;

	spotchk = 0;
	if(cert)
		syslog(0, HTTPLOG, "httpd starting on %s using %s", address, certificate);
	else
		syslog(0, HTTPLOG, "httpd starting on %s", address);
	ctl = announce(address, dir);
	if(ctl < 0){
		syslog(0, HTTPLOG, "error: can't announce on %s: %r", address);
		return;
	}
	//syslog(0,HTTPLOG,"%s %s", address, dir);
	/* 	then
	*	address is "tcp!*!80"
	*	dir is " /net/tcp/23"
	*	term% netstat -n
	*	tcp  23   web        Listen       80         0          ::
	*/
	strcpy(netdirb, dir);
	p = nil;
	if(netdir[0] == '/'){
		p = strchr(netdirb+1, '/');
		if(p != nil)
			*p = '\0';
	}
	if(p == nil)
		strcpy(netdirb, "/net");
	netdir = netdirb; // this is "/net"

	t0 = time(nil);
	k = 0;
	for(;;){
		/*
		 *  wait for a call (or an error)
		 */
		nctl = listen(dir, ndir);
		/*	example:
		*	dir="/net/tcp/23"	# listenning using this directory
		*	ndir="/net/tcp/33"	# new connection
		*	dir is a listenning directory
		*	ndir is the assigned directory
		*	nctl is opened in listen() by open("/net/tcp/23/listen", ORDWR)
		*	and should be closed in signle connection
		*	listen() reads "/net/tcp/23/listen" to find ndir
		*	listen() returns open fd of "/net/tcp/23/listen"
		*	Ref: /sys/src/libc/9sys/announce.c		*/
		if(nctl < 0){
			syslog(0, HTTPLOG, "error: can't listen on %s: %r", address);
			syslog(0, HTTPLOG, "error+ ctls = %d", ctl);
			return;
		}

		/*	Some protection against DoS attack that intends to make
		*	huge numbers of connected port.
		*	Note that persistent connection loop cannot be detected here.
		*	Be careful not to reject requests from proxy server.	*/

		ip = remoteaddr(ndir);
		puthlist(ip);
		//syslog(0,HTTPLOG,"%s call %s", ip, port);
		if(black(ip)){
			creject(ctl, ndir);
			close(nctl);
			continue;
		}

		/*
		*	int   accept(int ctl, char *dir);
		*	int   reject(int ctl, char *dir, char *cause);
		*	accept() accepts a call received by listen(),
		*	while reject() refuses the call because of cause. 
		*	Ref: dial(2)
		*/
		t = time(nil);
		//syslog(0,HTTPLOG,"time: %d %d %d %d %d", t, t0, k, obstime,maxaccess);
		/*	time(nil) returns local clock in starting machine.
		*	however in few 10 seconds it might change to the time based on NTP.
		*	then time returned by time(nil) will jump discontinuously.
		*	the condition t >= t0 is added to protect the anomaly that comes
		*	from this phenomenon	*/
		if(t >= t0 && t - t0 < obstime)
			k++;
		else{
			k = 0;
			t0 = t;
		}
		if(k > maxaccess){
			/*	The condition is system dependent,
			*	therefore required tuning.	*/
			creject(ctl, ndir);
			close(nctl);
			ip = arrest(obstime); // detect the ip that causes busy
			if(ip){
				putblist(ip); // and throw that into blacklist
				syslog(0,"blacklist", "%s %s # maxaccess", port, ip);
			}
			else // should not happen
				syslog(0,"blacklist", "%s unknown # maxaccess", port);
			continue;
		}
	
		n = nconnect(ip, &m);
		if( n > maxconnect){
			syslog(0,"blacklist", "%s %s # maxconnect", port, ip);
			putblist(ip);// throw the ip into blacklist
			creject(ctl, ndir);
			close(nctl);
			continue;
		}
		if( m > connectlimit){
			creject(ctl, ndir);
			close(nctl);
			continue;
		}

		/*
		 *  start a process for the service
		 */
		switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
		case -1:
			close(nctl);
			continue;
		case 0:
			/*
			 *  see if we know the service requested
			*
			*	What is the first argument of accept() should be ?
			*	man dial(2) says, it is nctl
			*	however /sys/src/cmd/ip/httpd/httpd.c says, ctl
			 */
			data = accept(ctl, ndir);
			if(data < 0){
				syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
				exits(nil);
			}
			if(cert){
				memset(&conn, 0, sizeof(conn));
				conn.cert = cert;
				conn.certlen = certlen;
				data = tlsServer(data, &conn);
				if(data < 0){// added by Kenar
					syslog(0, HTTPLOG, "error: tlsServer certificate=%s: %r", certificate);
					exits(nil);
				}
			}
			dup(data, 0);
			dup(data, 1);
			dup(data, 2);
			close(data);
			close(ctl);
			close(nctl);

			if(!usermode && certificate)
				becomenone();
			doservice(ndir);

			exits(nil);
		default:
			close(nctl);
			break;
		}

		if(++spotchk > 50){
			spotchk = 0;
			contentinit();
			urlinit();
			statsinit();
		}
	}
}

static void
doservice(char *ndir)
{
	HSPriv *hp;
	HConnect *c;
	NetConnInfo *nci;
	int ok;
	char *uri, *p;

	nci = getnetconninfo(ndir, -1);
	c = mkconnect();
	hp = mkhspriv();
	hp->remotesys = nci->rsys;
	hp->remoteserv = nci->rserv;
	myip = nci->lsys;
	port=nci->lserv; // this is real port
	c->private = hp;

	hinit(&c->hin, 0, Hread);
	hinit(&c->hout, 1, Hwrite);

	/*
	 * original comment for timeout is here:
	 * 	serve requests until a magic request.
	 * 	later requests have to come quickly.
	 * 	only works for http/1.1 or later.
	 * I don't know the meaning. the code was:
	 *	for(timeout = 15*60*1000; ; timeout = 15*1000){
	 * I replaced the code by simple one.
	 * -Kenar-
	 */

	/*	It's safe to prevent redirection loop here	*/
	for(;;){
		if(hparsereq(c, alivetimeout*1000) <= 0) // keep alive timeout (in unit of milisec)
			exits(nil);
		if(ntossed > 3){ // attack or redirection loop.
			/*	3 may be sufficient for allowed toss count	*/
			syslog(0, HTTPLOG, "warning: redirection loop ?");
			hfail(c, HInternal);
			exits(nil);
		}

		/*
		*	hparsereq() get basic parameters in the request:
		*	c->req.meth		# method "GET", "POST" ...
		*	c->req.vermaj
		*	c->req.vermin
		*	c->req.urihost		# nil, the value will be assigned by hparseheaders()
		*	c->req.uri		# uripath	got from GET line
		*	c->req.options		# ;params?query
		*	c->req.params		# params	got from GET line
		*	c->req.search		# query	got from GET line
		*
		*	scheme://urihost:port/uripath;params?search
		*	we should use real port no.
		*
		*	and also we have:
		*	c->req.user	nil
		*	c->req.path	nil
		*	these value will be assigned next.
		*	Note that there is no scheme info in the request.
		*	-Kenar-
		*/
		c->req.user = nil;
		uri = c->req.uri;
		c->req.path = uri;
		if(uri[0] == '/' && uri[1] == '~'){
			p = strchr(uri+2,'/');
			if(p)
				*p = 0;
			c->req.user = hstrdup(c, uri+2);
			if(p)
				*p = '/';
			c->req.path = p;
		}
		/*
		*	if uripath is "/~alice/foo", then
		*	c->req.user	"alice"
		*	c->req.path	"/foo"
		*	else if uripath is "/foo", then
		*	c->req.user	nil
		*	c-req.path	"/foo"
		*/

		ok = doreq(c);

		if(c->head.closeit || ok < 0)
			exits(nil);

		hreqcleanup(c);
	}
}


static int
doreq(HConnect *c)
{
	char *uri, *origuri, *p, *host, *user;
	char virtualhost[100];
	static char *host0 = nil;
	static char *user0 = nil;
	char *webroot = nil;
	char buf[256];
	int status;
	int n;

	uri = c->req.uri;
	/*
	 * normal case is just file transfer
	 */
	if(cont > 0){
		logit(c, "continue: %d",cont);
		if(cont > contmax){ /* may be required for safty */
			logit(c, "continuation exceed max: %d uri=%s",cont, uri);
			return hfail(c, HInternal);
		}
		cont++;
	}

	/*
	*	Kenar complains:
	*	if(hparseheaders(c, 15*60*1000) < 0)
	*	why so long time for parsing headers?
	*	I replaced by:
	*/
	if(hparseheaders(c, parsetimeout*1000) < 0)
		exits("failed");
	/*
	*	for old client c->head.host may be nil
	*	we must supply the host name from ip value
	*/
	if(c->head.host == nil)
		c->head.host = hstrdup(c, getdom("ip", myip));
	if(c->head.host == nil){
		logit(c, "host: nil");
		exits("failed");
	}
	/*
	*	now we will have:
	*	c->head.host
	*		if HTTP Host header exists, the value will be assigned from this,
	*		if not mydom is assigned.
	*		the value is already lowered.
	*	Let ar be a host name and aichi-u.ac.jp is the dom
	*	then head.host is something like:
	*	ar
	*	ar.aichi-u.ac.jp
	*	ar.aichi-u.ac.jp:80	# some browsers send HTTP Host header with port number
	*
	*	c->head.host gives mydom info for uriconnection
	*	we can establish uriconnection
	*/
	host = c->head.host;
	n = strlen(scheme) + strlen(host) + 10;
	uriconnection = halloc(c, n);
	/*	in this stage port part exist in host	*/
	snprint(uriconnection, n, "%s://%s", scheme, host);

	/*
	*	Let's remove ":port" here from urihost
	*	although this should be done in library
	*	-Kenar-
	*/
	host = hstrdup(c, host);
	p = strchr(host,':');
	if(p)
		*p = 0;

	if(!http11(c) && !c->head.persist)
		c->head.closeit = 1;
	logit(c, "%s %s%s", c->req.meth, uri, c->req.options);

	/*
	*	we can remove padom part from host
	*	so that we can simplify management of /sys/lib/httpd.rewrite
	*	now we don't need full name for virtual host in this file.
	*/
	if(strlen(host)>strlen(padom)){
		char *p, buf[256];
		snprint(buf, sizeof buf, ".%s", padom);
		p = host + strlen(host) - strlen(buf);
		if(strcmp(p, buf) == 0)
			*p = 0;
	}
	/*	then host is "ar in ar.aichi-u.ac.jp" for example
	*	on the other hand, host "202.250.160.40" remains as it is	*/

	//logit(c, "referer:%s", c->head.referer);

	/*
	*	We should examine the connection is really continued ?
	*	How to determine the connection should enter new name space?
	*	How to enter new name space?
	*	HTTP referer defined in HTTP/1.0 and HTTP/1.1 may be helpful.
	*	referer is http://host/path if URI is from href in html document
	*	example:	http://pc/~alice/photos/
	*	According to RFC, referer can be set not to be send for privacy
	*	But referer is nil if URI is from keyboard input,
	*	therefore referer is not acceptable for this purpose.
	*	
	*	We claim to resend if (host, user) pair has changed
	*	and close the connection
	*
	*	schemes and ports concern nothing to this problem.
	*	because different servers care for them
	*/
	user = c->req.user;
	if(cont){
		if(xstrcmp(host, host0) != 0 || xstrcmp(user, user0) != 0){
			/*	claim to resend and close the connection */
			c->head.closeit = 1;
			hmoved(c, c->req.uri);
			logit(c,"resend to:  %s%s", uriconnection, c->req.uri);
			return -1; //exit
		}
	}
	else{
		host0 = estrdup(host);
		if(user) user0 = estrdup(user);
		/*	don't use hstrdup because memory is initialized
		*	in continued connection		*/
	}

	if((strcmp(scheme,"https") == 0 && strcmp(port,httpsport) == 0)
	|| (strcmp(scheme,"http") == 0 && strcmp(port,httpport) == 0))
		snprint(virtualhost, sizeof virtualhost, "%s://%s/", scheme, host);
	else
		snprint(virtualhost, sizeof virtualhost, "%s://%s:%s/", scheme, host, port);
	origuri = c->req.uri;
	uri = nil;
	if(user == nil)
		uri = redirect(c, rewrite, virtualhost, &webroot);
	#ifdef NOTE
	*	we have three cases of uri
	*	1. 	scheme://host/foo
	*	2. 	/foo
	*	3. 	nil
	#endif
	if(uri){
		/*	Incoming request is limited by /sys/lib/httpd.rewrite to our host	*/
		/*
		if(uri[0] == '*'){
			webroot = hstrdup(c, uri + 1);
			binding = 1;
		}
		*/
		if(webroot)
			binding = 1;
	}
	else{
		/*
		*	we can reject some faked accesses here;
		*	if uri is nil and host is not myname, we can reject.
		*	example:
		*	ar	# sysname
		*	plan9	# virtual host of ar. To be rejected.
		*	aichi-u.ac.jp	# padom
		*	now ar.aichi-u.ac.jp and plan9.aichi-u.ac.jp are reduced to
		*	ar and plan9 respectively.
		*	On the other hand
		*	www.faked.com
		*	is kept untached.
		*/
		if(strcmp(host, myname) != 0 && strcmp(host, myip) != 0){
			if(user)
				logit(c, "denied: %s/~%s", host, user);
			else
				logit(c, "denied: %s", host);
			return hfail(c, HNotFound);
		}
		/*	redirection is examined only for real host	*/
		uri = redirect(c, rewrite, origuri, &webroot);
		//logit(c,"uri=%s origuri=%s webroot=%s", uri, origuri, webroot);
		if(uri && webroot){
			/*	we just accept the incoming request	*/
			c->req.path=uri;
			if(c->req.user)
				binding = 2;
			else
				binding = 1;
		}

	}

	mnt = 0;
	if(webroot){
		#ifdef NOTE
			This case is '*' redirect in /sys/lib/httpd.rewrite such as:
			http://pegasus  */usr/someone/www
			/~alice	*/usr/alice/www
			/car		*/usr/carol/www
			/		*/usr/bob/www
		#endif
		if(!cont){
			status = bind(webroot, webmnt , MREPL);
			if(status < 0){
				logit(c,"error: bind %s %s # %r", webroot,webmnt);
				return hfail(c, HNotFound);
			}
			mnt = 1;
		}
	}
	else if(uri != nil){ /* Let the client know the new address */
		char *nu;
		nu = addoptions(c, uri);
		logit(c,"redirect1 to:  %s ntossed=%d", nu, ntossed);
		ntossed++;
		return hmoved(c, nu);
	}


	/*
	* 	Here we support http://ar/~alice/abc
	*	where ar must be real host and need noweb == 0.
	*	Then we bind /usr/alice/web to /usr/web/mnt
	*	and change request document to /abc
	*	This redirection is automatically done without rewrite-file
	*	-Kenar-
	*/
	uri = c->req.uri;
	if(noweb == 0 && c->req.user && binding != 1){
		/*	then uri[0]=='/' && uri[1]=='~'	*/
		char *t;
		/* 	if uri is "/~alice" for example,
			let the client resend the correct uri "/~alice/"
			-Kenar-	*/
		t = strchr(uri+2, '/');
		if(!t){
			snprint(buf, sizeof buf, "%s/", uri);
			logit(c, "redirect2 to: %s", buf);
			return hmoved(c, buf);
		}

		if(!cont && !mnt){
			char *user;
			char *path;
			int n;
			user = c->req.user;
			n = strlen("/usr//web") + strlen(user) + 1 + UTFmax;
			path = halloc(c, n);
			snprint(path, n, "/usr/%s/web", user);
			status = bind(path, webmnt, MREPL);
			if(status == -1){
				logit(c, "error: bind %s %s # %r", path, webmnt);
				return notfound(c, uri+1);
			}
			mnt = 1;
		}

		/* 	then c->req.uri is "/~alice/abc", and path is "/abc" 
			for the request http://ar/~alice/abc
			We need not be wary of "/~alice/../",
			it reduces to "/" as it should be.
			-Kenar-	*/
	}

	/* namespace configuration; hide everything out of webbase */
	if(!cont){
		if(addns(runas, namespace) < 0){
			logit(c, "error: addns %s # %r", namespace);
			sysfatal("can't build httpd namespace");
		}
		if(mnt){
			char nsbuf[16];
			/* 	strlen("namespace_65536") is 15	*/
			if(nsconfig(c, webbase) < 0){
				logit(c, "error: nsconfig %s # %r", webbase);
				return -1;
			}
			snprint(nsbuf, sizeof nsbuf, "/etc/namespace_%s", port);
			//logit(c,"etc/namespace: %s", nsbuf);
			if(mntok)
				addns(runas, nsbuf);
			else
				chgns(nsbuf);
		}
		cont = 1;
	}
	chdir("/");

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	return send(c);
}

/*	# return values
*	+1: OK
*	-1: NG
*	-Kenar-	*/
static int
send(HConnect *c)
{
	Dir *dir;
	Handle *h;
	char *p, *abspath;
	char *uri, *newuri;
	char *path;
	int n, ok;
	ulong dirmode;
	char buf[256];
	int status;
	HSPriv *hp = c->private;

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	if(strcmp(c->req.meth, "GET") != 0
		&& strcmp(c->req.meth, "HEAD") != 0
		&& strcmp(c->req.meth,"POST") != 0)
		return hunallowed(c, "GET, HEAD, POST");
	/*	NOTE
	*	PUT and DELETE method are supported by
	*		Mozilla/4.5
	*	but not supported by
	*		IE/6.0 nor Mozilla/6.2
	*	so it may be worthless to support these method
	*	-Kenar-	*/

	if(c->head.expectother || c->head.expectcont)
		return hfail(c, HExpectFail);

	path = c->req.path;
	uri = c->req.uri;

	newuri = redirect(c, "/etc/rewrite", path, nil);
	if(newuri){
		char *nu;
		nu = addoptions(c, newuri);
		logit(c, "moved to: %s", nu);
		ntossed++;
		return hmoved(c, nu);
	}

	/*
	 * check for directory/file mismatch with trailing /,
	 * and send any redirections.
	 */
	n = strlen("/doc") + strlen(path) + strlen("/index.html") + 10;
	/* make a room to insert index.html */
	abspath = halloc(c, n);
	strcpy(abspath, "/doc");
	strcat(abspath,path);

	/*
	*	we examine if it is a directory
	*	note that the path need not exist
	*	it may be handled by handler
	*/
	if((dir = dirstat(abspath)) != nil){
		dirmode = dir->mode;
		free(dir);
		if(dirmode & DMDIR){
			/*
			*	if path does not end with '/'
			*	then let the client resent the uri with "/"
			*	Don't replace by the code such as the following
			*	because HTTP/1.0 client will not work
			*	They hate redirection
			*	-Kenar-
			*
			* 	if(path[strlen(path)-1] == '/')
			*		snprint(buf, sizeof buf, "%sindex.html", uri);
			* 	else
			*		snprint(buf, sizeof buf, "%s/index.html", uri);
			* 	return hmoved(c, buf);
			*/
			if(uri[strlen(uri)-1] != '/'){
				snprint(buf, sizeof buf, "%s/", uri);
				logit(c, "redirect3 to: %s", buf);
				return hmoved(c, buf);
			}
			/*
			*	here the form of uri is "path/"
			*	that is, path ends with '/'
			*	-Kenar-
			*/
			strcat(abspath, "index.html");
		}else if((p = strchr(abspath, '\0')) > abspath && p[-1] == '/'){
			/*
			*	This condition means:
			*	path is a file and ends with '/' 
			*	for example,
			*		http://ar.aichi-u.ac.jp/index.html/
			*	Then let the client to resend
			*		http://ar.aichi-u.ac.jp/index.html
			*	-Kenar-
			*/
			*strrchr(uri, '/') = '\0';
			logit(c, "%s hmoved", uri);
			return hmoved(c, uri);
		}
	}


	//	--------------------------
	//	req.uri and req.path are redefined
	//	--------------------------

	/*	Let our sendit() know the contents type to assign full pathname
	*	to req.uri
	*	I dislike to replace req.uri, however sendit() uses file suffix in
	*	req.uri to determine file type.
	*	-Kenar-	*/
	path = abspath + 4; // 4 is strlen("/doc"), now `path' is path to a file
	c->req.path = path; // file path from document root

	/*
	*	Now, variables
	*		path abspath
	*	are path to file
	*/

	/*
	*	access control must be the last
	*	because it is delicate to handle last slash
	*	so full path is required to evaluate
	*	-Kenar-	*/

	if(ipcheck(c, "/etc/allow", hp->remotesys, path) < 0){
		logit(c, "ipcheck: unallowed %s %s", hp->remotesys, path);
		return hfail(c, HUnauth, path);
	}

	ok = authcheck(c, "/etc/passwd", path);
	if(ok <= 0){
		logit(c, "authcheck: unallowed %s", path);
		return ok;
	}

	/*	Now every redirection is finished and we have ready.
	*	requested path is already converted to absolute path.
	*	values to be used are:
	*	path: path to the file to send. where path is in httpd name space
	*	fd: open file descriptor of path
	*
	* 	CGI is supported using exechandler.
	*	The handler configuration file is /etc/handler.
	*	Handler will be obtained from requested path.
	*	Handler must be exec permitted.
	*	-Kenar-
	*/

	/* we disallow files starting with `.' */
	p = strrchr(abspath,'/');
	if(p[1] == '.'){
		logit(c, "%s not found", uri);
		return notfound(c, uri);
	}

	ntossed = 0;

	/*	call exechandler	*/
	h = handler(c, "/etc/handler", path);
	if(h){
		char **argv;
		int n;
		argv = h->argv;
		for(n = 0; argv[n]; n++)
			if(strcmp(argv[n],"$target") == 0)
				argv[n] = abspath;
		if(*argv[0] != '/'){
			logit(c, "/etc/handler: %s must be absolute path", argv[0]);
			return hfail(c, HInternal);
		}
		logit(c,"%s", argv[0]);
		if(executable(argv[0])){
			/*	Don't use access(path, AEXEC)==0 instead of executable(path)
			*	Why? If the fs is running in allow mode, all files are
			*		access(path, AEXEC)==0
			*	-Kenar-	*/
			if(access("/unreadable", AREAD) == 0){
				logit(c,"in mentenance");
				return hfail(c, HUnabailable);
			}

			if((strcmp(c->req.meth,"POST") == 0) && (acceptpost(c, posttimeout) < 0))
				return hfail(c, HReqSize);

			/* chdir to the target dir, taget is abspath in httpd space*/
			p = strrchr(abspath,'/');
			*p = 0;
			if(chdir(abspath) < 0)
				chdir("/doc");
			*p = '/';

			setcgienv(c);
			ramfs(c);
			status = execute(c,argv, h->mimetype);
			unmount(nil, "/tmp");
			return status;
		}
	}
	status = sendit(c, abspath, nil, nil);
	return status; // +1: success;  -1: fail
}

static	char*
sysdom(void)
{
	char *dn;

	dn = csquery("sys" , sysname(), "dom");
	if(dn == nil)
		dn = "who cares";
	return dn;
}

/*
 *  query the connection server
 */
static	char*
csquery(char *attr, char *val, char *rattr)
{
	char token[64+4];
	char buf[256], *p, *sp;
	int fd, n;

	if(val == nil || val[0] == 0)
		return nil;
	snprint(buf, sizeof(buf), "%s/cs", netdir);
	fd = open(buf, ORDWR);
	if(fd < 0)
		return nil;
	fprint(fd, "!%s=%s", attr, val);
	seek(fd, 0, 0);
	snprint(token, sizeof(token), "%s=", rattr);
	for(;;){
		n = read(fd, buf, sizeof(buf)-1);
		if(n <= 0)
			break;
		buf[n] = 0;
		p = strstr(buf, token);
		if(p != nil && (p == buf || *(p-1) == 0)){
			close(fd);
			sp = strchr(p, ' ');
			if(sp)
				*sp = 0;
			p = strchr(p, '=');
			if(p == nil)
				return nil;
			return estrdup(p+1);
		}
	}
	close(fd);
	return nil;
}

/*------- Kenar's addition follows --------*/

/*
*	get parent domain of the server	*/
static	char *
pardom(void)
{
	char *sn, *sd, *pd;
	sn = sysname();
	sd = sysdom();
	if(strlen(sd) < strlen(sn)) // something wrong
		return nil;
	pd = sd + strlen(sn) + 1; // parent dom
	return pd;
}


static void
setcgienv(HConnect *c)
{
	HSPriv *hp = c->private;
	char buf[256];
	char *cputype, *name;
	int ch;

	cputype = getenv("cputype");

	rfork(RFCENVG);

	/*	Putting header info into environment
	*	must be the first. (for security)
	*	-Kenar-	*/
	ch = *(c->hstop);
	*(c->hstop) = 0;
	setenv("HTTP_HEADER", (char*)c->header);
	hparse2env(c, (char*)c->header);
	*(c->hstop) = ch;

	setenv("DOCUMENT_ROOT", "/doc");
	setenv("GATEWAY_INTERFACE", "CGI/1.1");
	snprint(buf, sizeof buf, "%s!%s", hp->remotesys, hp->remoteserv);
	setenv("REMOTE_ADDR", buf);

	setenv("REQUEST_PATH", c->req.path);
	setenv("REQUEST_METHOD",c->req.meth);
	
	snprint(buf, sizeof buf, "%d.%d", c->req.vermaj,c->req.vermin);
	setenv("REQUEST_VERSION", buf);
	setenv("REQUEST_URI",c->req.uri);
	setenv("AUTH_USER", c->head.authuser);
	setenv("AUTH_PASS", c->head.authpass);
	setenv("REQUEST_USER", c->req.user);

	setenv("SERVER_DOMAIN", mydom);
	setenv("SERVER_NAME", sysname());
	setenv("SERVER_PORT", port);
	setenv("SERVER_PROTOCOL",hversion);
	setenv("SERVER_SOFTWARE", serversoft);
	setenv("QUERY_STRING", c->req.search);
	setenv("target", c->req.path);
	name = strrchr(c->req.path, '/');
	setenv("name", name+1);
	setenv("query", c->req.search);
	snprint(buf, sizeof buf, "%D", time(nil));
	setenv("date", buf);
	setenv("home", "/doc");
	setenv("cputype", cputype);
	setenv("objtype", cputype);
}

/*
*	configure service namespace
*/
static int
nsconfig(HConnect *c, char *webbase)
{
	char *wmnt, *wdoc, *wetc, *wbin, *wmdoc, *wmetc, *wmbinrc, *wmbinobj;
	wmnt = hcat2(c, webbase,"/mnt");
	wdoc = hcat2(c, webbase, "/doc");
	wetc = hcat2(c, webbase, "/etc");
	wbin = hcat2(c, webbase,"/bin");
	wmdoc = hcat2(c, webbase, "/mnt/doc");
	wmetc = hcat2(c, webbase, "/mnt/etc");
	wmbinrc = hcat2(c, webbase, "/mnt/bin/rc");
	wmbinobj = hcat3(c, webbase, "/mnt/bin/",getenv("objtype"));
	/*	 the two unmounts are for the guard
	*	when next two binds fail.
	*	for example, /usr/web/mnt/etc is absent	*/
	unmount(nil, wdoc);
	unmount(nil, wetc);
	if(bind(wmdoc, wdoc, MREPL) < 0)
		return -1;
	bind(wmetc, wetc, MREPL);
	bind(wmbinrc, wbin, MAFTER);
	bind(wmbinobj, wbin, MAFTER);
	unmount(nil, wmnt);
	bind(webbase,"/", MREPL);
	return 0;
}


static int
acceptpost(HConnect *c, int timeout)
{
	int fd[2];
	char *hb;
	Hio *hin;
	ChunkBuf *cb, *cb0, *cb1;
	int m,n;

	alarm(timeout*1000);

	/*	DEBUG for receiving post data over 32KB
	*	Ref: echo.c	
	print("%s 200 OK\r\n", hversion);
	print("Server: %s\r\n", serversoft);
	print("Date: %D\r\n", time(nil));
	print("Content-Type: text/plain\r\n");
	print("Connection: close\r\n");
	print("\r\n");
	hin = hbodypush(&c->hin, c->head.contlen, c->head.transenc);
	if(hin == nil){
		logit(c, "error: nil hbodypush");
		hfail(c, HInternal);
		exits("failed");
	}
	for(; hb = hreadbuf(hin, hin->pos); hin->pos = hin->stop)
		write(1, hb, hbuflen(hin, hb));
	exits(nil);
	*/

	if(c->head.contlen > maxpost){
		logit(c, "error: post data too large: %d", c->head.contlen);
		hfail(c, HReqSize, maxpost);
		exits("failed");
	}
	/*	head.contlen can be faiked. so real protection is needed	*/
	hin = hbodypush(&c->hin, c->head.contlen, c->head.transenc);
	if(hin == nil){
		logit(c, "error: nil hbodypush");
		hfail(c, HInternal);
		exits("failed");
	}
	m = 0;
	cb0 = cb = nil;
	for(; hb = hreadbuf(hin, hin->pos); hin->pos = hin->stop){
		cb1 = malloc(sizeof *cb1);
		if(cb1 == nil){
			logit(c, "error: nil malloc ChunkBuf");
			hfail(c, HInternal);
			exits("failed");
		}
		n = hbuflen(hin, hb);
		memcpy(cb1->buf, hb, n);
		cb1->size = n;
		m += n;
		if(m > maxpost){
			logit(c, "error: post data too large: %d", c->head.contlen);
			hfail(c, HReqSize, maxpost);
			exits("failed");
		}
		if(cb0 == nil)
			cb0 = cb1;
		if(cb != nil)
			cb->next = cb1;
		cb = cb1;
	}
	if(cb == nil){
		hfail(c, HBadReq);
		exits("failed");
	}
	cb->next = nil;
	alarm(0);

	/* pass the hb to exec() */
	pipe(fd);
	switch(fork()){
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		/*	Let the child exit, because I would like to let CGI
		*	succeed to this pid
		*	-Kenar-	*/
		close(0);
		close(fd[0]);
		dup(fd[1], 1);
		close(fd[1]);
		/*	POST data have been received in ChunkBuf cb0,
		*	so we pass the data to the child process */

		cb = cb0;
		while(cb){
			m=write(1, cb->buf, cb->size);
			if(m != cb->size){
				logit(c, "error: write # %r");
				hfail(c, HInternal);
			}
			cb = cb->next;
		}
		exits(nil);
	default:
		close(fd[1]);
		dup(fd[0], 0);
		close(fd[0]);
		/* remove ChunkBuf pointed by cb0 */
		cb = cb0;
		while(cb){
			cb1 = cb->next;
			free(cb);
			cb = cb1;
		}
		break;
	}
	return 0;
}
			
static int
execute(HConnect *c, char *hargv[], char *mimetype)
{
	char *p, *q;
	char *params;
	int k, n, hargc;
	char **argv;
	char *path;

	params = c->req.params;
	/*	params points "alice;bob"
	*	In creating argv for exec call,
	*	we should allocate memory for argv.
	*	argv[0] must be reserved for basename of exec file
	*	argv[n-1] must be nil where n is num of elements
	*	so, I use ";alice;bob;" to apply getfield.
	*	-Kenar-		*/
	if(params == nil)
		argv = hargv;
	else{
		for(n = 0; hargv[n]; n++);
		hargc = n;
			
		n = strlen(";") + strlen(params) + 1 + UTFmax;
		q = halloc(c, n);
		snprint(q, n, "%s;", params);
		params = q;
		n = 0;
		for(p = params; *p; p++)
			if(*p == ';') n++;
		argv = halloc(c, (hargc + n+1) * sizeof(char*));
		n = getfields(params, &argv[hargc], n+1, 0, ";");
		for(k = 0; k < hargc; k++)
			argv[k] = hargv[k];
		/* don't forget to hurlunesc */
		for(k = hargc; k < hargc + n -1; k++)
			argv[k] = hurlunesc(c, argv[k]);
		/* it is not good idea to trim extra spaces */
		argv[hargc + n-1] = nil;
	}
	path = argv[0];
	p = strrchr(argv[0],'/');
	argv[0] = p + 1;

	/*	if it takes so much time to complete, kill the process
	*	-Kenar-	*/
	return execsafely(c, path, argv, mimetype);
}


/*	This code intends to get exit status from the child	*/
static int
execsafely(HConnect *c, char *path, char **argv, char *mimetype)
{
	int fd, n;
	int status;
	char msg[ERRLEN];
	char *exitmsg;
	Dir *dir;
	char buf[4*1024];

	/* 	wait child to finish.
	*	If it takes too much time, kill the child and get message  */
	switch(cpid = rfork(RFFDG|RFPROC|RFNOTEG|RFREND|RFNAMEG)){ /* assign = */
	/*	RFNAMEG: the new process inherits a copy of the parent's name space.
	*	RFNAMEG is required so that nomnt() does not make effect parent namespace	*/
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0: // child: read stdin, and write to stdout
		fd = create("/tmp/...", OWRITE,0660);
		if(fd < 0){
			logit(c, "/tmp/... unable to create");
			exits("500 Internal Server Error");
		}
		dup(fd, 1); // the stdout is redirected to fd
		dup(fd, 2); // the stderr is also redirected to fd
		close(fd);
		if(!mntok)
			nomount(c);
		exec(path, argv);
		/* not reached */
		logit(c,"error: exec %s # %r", path);
		exits("500 Internal Server Error");// don't return
		break;
	default:// parent: wait for child to finish.
		break;
	}
	fd = dup(0, -1);
	close(0);
	atnotify(notefun,1);
	alarm(cgitimeout*1000);
	notestatus = 1;
	status = waitfor(cpid, msg);
	alarm(0);
	atnotify(notefun,0);
	dup(fd, 0);
	close(fd);
	if(notestatus != 1){
		logit(c,"alarm done to pid=%d", cpid);
		return hfail(c, HInternal);
	}
	if(status < 0){
		logit(c,"error: no children to wait %d", cpid);
		return hfail(c, HInternal);
	}

	/*	example of msg:
	*	httpd 1653: 500 Internal Server Error
	*	therefore we must extract what we want	*/
	exitmsg = nil;
	if(msg[0] != 0){
		exitmsg = strchr(msg, ':');
		if(exitmsg && exitmsg[1] == ' ')
			exitmsg += 2;
		else
			exitmsg = msg;
		logit(c,"child %d exited with msg [%s]", cpid, exitmsg);
	}
	fd = open("/tmp/...", OREAD);
	if(fd< 0){
		logit(c, "/tmp/... not open");
		return hfail(c, HInternal);
	}

	status = -1;
	if(strcmp(mimetype, "-") != 0){
		dir = dirfstat(fd);
		if(dir == nil){
			logit(c, "dirfsstat of /tmp/... nil");
			close(fd);
			return hfail(c, HInternal);
		}
		printheaders(c, exitmsg, mimetype, dir->length);
		free(dir);
		status = 1;
	}

	while((n = read(fd,buf, sizeof buf)) > 0)
		write(1, buf, n);
	close(fd);
	if(exitmsg)
		logit(c, "done %s # [%s]",path, exitmsg);
	else
		logit(c, "done %s", path);
	return status;
}

static int
ramfs(HConnect *c)
{
	int pid, status;
	char msg[ERRLEN];
	/*	ramfs service to the child	*/
	switch(pid = rfork(RFFDG|RFPROC)){ /* assign = */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		close(0);
		close(1);
		close(2);// prevent "ramfs: unmount read" message
		execl("/bin/ramfs","ramfs","-p",0);
		/* not reached */
		logit(c, "error: unable to execl /bin/ramfs %r");
		exits(nil);
	default:
		break;
	}
	
	/* ensure "/bin/ramfs" finished */
	status = waitfor(pid, msg);
	if(status < 0){
		logit(c,"error: /bin/ramfs: no children to wait %d", pid);
		return -1;
	}
	return 0;
}

static int
nomount(HConnect *c)
{
	int status;
	status = rfork(RFNOMNT);
	if(status < 0){
		logit(c,"error: rfork nomnt # %r");
		return hfail(c, HInternal);
	}
	return 0;
}

/*	mime parse is not easy.
*	and probably meaningless to put the info int env.
*	so I chose a simple way
*	I parsed only
*		key:val
*	pairs, and
*		putenv(key, val)
*	-Kenar-	*/
static int
hparse2env(HConnect *c, char *head)
{
	char *p,*s;
	char **buf;
	char *kv[2];
	char buf1[64];
	int k,n,m;

	p = hstrdup(c, head);
	/* skip a line */
	p = strchr(p, '\n');
	if(!p)
		return -1;
	p++;
	n = 0;
	for(s = p; *s; s++)
		if(*s == '\n') n++;
	buf = halloc(c, (n + 1) * sizeof(char*));
	n = getfields(p, buf, n + 1, 0, "\n");
	
	/* strip the trailing CR */
	for(k = 0; k < n; k++){
		p = strchr(buf[k], '\r');
		if(p)
			*p = 0;
	}

	/*	get
	*		key: val
	*	and trim, then putenv	*/
	for(k = 0; k < n; k++){
		m = getfields(buf[k], kv, 2, 0, ":");
		if(m != 2)
			continue;
		for(p = kv[0]; *p; p++)
			*p = toupper(*p);
		for(p = kv[0]; *p; p++)
			if(*p == '-') *p = '_';
		snprint(buf1, sizeof buf1, "HTTP_%s", trim(kv[0]));
		setenv(buf1, trim(kv[1]));
	}
	return 0;
}


/*	"/proc" is required for postnote	*/

static int
notefun(void *a, char *msg)
{
	USED(a);
 	postnote(PNGROUP, cpid, msg);
	if(strcmp(msg, "alarm") == 0){
		notestatus = 0;
		return 1;
	}
	return 0;
}

/*
*	The format of "/etc/allow" is shown using example:
*	/lecture/for/our/students
*		192.168.[1-8].*
*		...
*	path
*		...
*	where ip is given by shell style pattern matching.
*
*	restrictions := restriction restriction ...
*	restriction := pathline iplines
*	pathline := path
*	iplines := ipline ipline ..
*	ipline := spaces ippattern ippattern ..
*
*	# return values
*	0: the IP is allowed
*	-1: the IP is disallowed
*/
static int
ipcheck(HConnect *c, char *file, char *ip, char *path)
{
	char *p, *line, *s;
	int ok, check;//, m;
	Biobuf *b;
	char word[64];

	//logit(c,"ipcheck: ip=%s", ip);

	b = Bopen(file, OREAD);
	if(b){
		ok = 1;
		check = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(strchr(" \t", line[0])){	// ip pattern of lpath
				if(check == 0) // skip the ip pattern
					continue;
				if(ok)
					continue;
				/* 	how can I get single pattern ?
				*	ip pattern has no space	OK */
				while(line = getword(line, word, sizeof word)){ // asign =
					//logit(c,"ipcheck: word=%s", word);
					if(word[0] == '~' && match(ip, word+1) == 0){
						ok = 1;
						break;
					}
					if(match(ip, word)){
						ok = 1;
						break;
					}
				}
				if(!ok) // examine more ip line
					continue;
			} else {	// path
				/* first we must check previous analysys */
				if(check && !ok)
					break;
				/* start new path line */
				p = strpbrk(line," \t");
				if(p)
					*p = 0;
				/* 	check some human error, in error put some
				*	guards and let that know. */
				if(line[0] != '/'){
					logit(c, "warning: path in /etc/allow must begin with '/': %s",line);
					path++;
				}
				if(deeper(path, line)){
					//logit(c, "deeper: %s %s",path,line);
					/* 	then we must confirm
					*	the client IP is allowed to see our data */
					check = 1;
					ok = 0;
				}
				else{
					check = 0;
					ok = 1;
				}
			}
		}
		Bterm(b);
		return  ok - 1;
	}
	return 0;
}

/*
*	uriloop(pat, rep, uri) returns:
*	 1 if rep makes loop from the pat
*	 0 if not
*
*	uri is incoming uri of common form
*		scheme://host[:port]
*	pat is one of the form:
*		scheme://host[:port][/path]
*		/path
*	rep is one of the form:
*		scheme://hostdom[:port][/path][;others]
*		/path[;others]
*	scheme and host of pat and rep is assumed to be lower.
*	port part is assumed to be eliminated if it is:
*		80 for http
*		433 for https
*/

static int
uriloop(char *pat, char *rep, char *uri)
{
	int c;
	char *p, *q;
	char *repport;
	//syslog(0,HTTPLOG,"uriloop: %s %s %s", pat, rep, uri);
	/*	comparison under common form is required	*/
	repport = nil;
	p = rep;
	if(strncmp(p, "http://", 7) == 0){
		p += 7;
		repport = httpport;
	}
	else if(strncmp(p, "https://", 8) == 0){
		p += 8;
		repport = httpsport;
	}
	if(repport == nil) // the rep is not this host, not looped
		return 0;
	/*	then p points host part	*/
	q = skipto(p, "/");
	c = *q;
	if(*q == '/')
		*q = 0;
	/*	then p is "ar.aichi-u.ac.jp" or "ar.aichi-u.ac.jp:8080" for example	*/
	if(strcmp(uri, rep) != 0){// the rep is not this host, not looped
		*q = c;
		return 0;
	}
	*q = c; // then q is path part that may be "/somepath" or ""
	if(*q) // then *q == '/'
		return deeper(q, pat);
	// then pat has not path part, we can asume 
	return (strlen(pat) == 1);
}

#ifdef NOTE
*	path redirection for users and virtual hosts
*	this redirection is used after namespace configuration
*
*	example:
*	request is http://ar/~alice/photo/index.html;args?query
*	then,
*	path is "/photo/index.html"
*	c->req.user is "alice"
*	c->head.host is "ar"
*	
*	How to design rewrite file ?
*	A line consists two fields except comment:
*		pattern	redirection
*
*	We support by this file:
*	1. local redirection
*	2. non local redirection
*	3. port redirection
*	4. scheme redirection
*	5. binding
*	We introduce special charactor '*'
*	then types of second fields are:
*		/path		# local redirection
*		*/path		# binding
*		scheme://host/path		# non local redirection
*		scheme://host:port/path	# non local redirection with port
*	- server must know current scheme and port to prevent redirection loop
*	but how to know current scheme?
*	- we have only two scheme that is accepted by httpd: http and https
*	the latter has certificate.
*	- it is enought to compare host address and port to prevent redirection loop.
*	scheme is needless.
*
*	Example redirection file that can be consulted by both http and https
*		http://vhost	*/vhostwebroot
*		/netlib	ftp://somehost			# note not only http or https
*		/oldplace	/newplace			# local redirection
*
*	Convert second field:
*		http://host	->	http://host:80
*	in checking redirection loop.
*	Pre-assign my scheme host port string such as:
*		http://ar:80
*	Then compare.
*	It is safe to remain scheme part so that we need not talk except http and https.
*
*	How should I do in case:
*		/		http://ar.aichi-u.ac.jp
*		/		http://202.250.160.40
*		/private	https://ar.aichi-u.ac.jp/private
*	These make a redirection loop if httpd in running on ar.aichi-u.ac.jp.
*	Hmm... "//*" is not good idea...
*	Second field requires full address.
*	Example redirection file that can be consulted by both http and https
*		http://vhost	*/vhostwebroot
*		/private	https://ar.aichi-u.ac.jp/private	# this is skipped by https
*		/special	http://ar.aichi-u.ac.jp:8080/special	# this is skipped by 8080 http
*		/netlib	ftp://somehost			# note not only http or https
*		/oldplace	/newplace			# local redirection
*		/		http://ar.aichi-u.ac.jp/	# this is skipped by http
*	
*
*	# return values
*	nil: not redirected
*	else: new uri
*	-Kenar-
#endif

static char*
redirect(HConnect *c, char *file, char *path, char **pwebroot)
{
	char *pat, *rep, *line, *s, *buf;
	Biobuf *b;
	char *field[2];
	int bufsize;
	char myuri[64]; // "http://ar.aichi-u.ac.jp/" ...  # scheme://host[:port]

	b = Bopen(file, OREAD);
	if(b){
		pat = nil;
		rep = nil;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			if(getfields(line, field, 2, 1, " \t") != 2)
				continue;
			if(deeper(path, field[0])){
				pat = field[0];
				rep = field[1];
				//logit(c, "redirect: deeper: %s %s", pat,rep);
				break;
			}
		}
		if(rep == nil){
			Bterm(b);
			return nil;
		}
		if(strcmp(pat,"/") == 0 && strncmp(path, "/~", 2) == 0){
			Bterm(b);
			return nil;
		}
		if(rep[0] == '*'){
			rep++;
			/*	we cut off trailing '/'	*/
			if(*(rep + strlen(rep) - 1) == '/')
				*(rep + strlen(rep) - 1) = 0;
			//rep = hstrdup(c,rep);	
			if(pwebroot)
				*pwebroot = hstrdup(c,rep);
			#ifdef NOTE
			*	In rewrite file we allow
			*	/car		*/usr/carol/www
			*	/~carol	*/usr/carol/www
			*	/		*/usr/bob/www
			* 	Then path="/car/somepath" should return "/somepath"
			#endif
			/*	cut off trailing "/" for safety. this is especially true for pat="/"	*/
			if(*(pat + strlen(pat) - 1) == '/')
				*(pat + strlen(pat) - 1) = 0;
			rep = hstrdup(c, path + strlen(pat));
		}
		else {// rep is "/...." or "scheme://...." 
			if((strcmp(scheme,"http") == 0 && strcmp(port,httpport) == 0)
			|| (strcmp(scheme,"https") == 0 && strcmp(port,httpsport) == 0))
				snprint(myuri, sizeof myuri, "%s://%s", scheme, c->head.host);
			else
				snprint(myuri, sizeof myuri, "%s://%s:%s", scheme, c->head.host,  port);
			if(uriloop(pat, rep, myuri)){
				//logit(c, "redirect: uriloop: 1");
				return nil;
			}
			bufsize = strlen(rep) + strlen(path) - strlen(pat) + 1 + UTFmax;
			buf = halloc(c, bufsize);
			s = path + strlen(pat);
			if(*s == '/' || *(rep+strlen(rep) -1) == '/' || *s == 0)
				snprint(buf, bufsize, "%s%s", rep, s);
			else
				snprint(buf, bufsize, "%s/%s", rep, s);
			rep = buf;
			if(pwebroot)
				*pwebroot = nil;
		}
		Bterm(b);
		return rep;
	}
	return nil;
}

/*	exec handler
*
*	# data format:
*	path	mimetype	ramfs	path	args ...
*
*	# example:
*	.dx_html	text/html	0	/bin/dx
*
*	# return values
*	nil: need not to handle
*	else: handling command line
*	-Kenar-	*/
static Handle*
handler(HConnect *c, char *file, char *path)
{
	char *line, *s, *p;
	Biobuf *b;
	char *field[16+1], **v;
	Handle *h;
	int n, ok;
	if(!path)
		return nil;
	b = Bopen(file, OREAD);
	if(b){
		ok = 0;
		p = halloc(c, 256);
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			strncpy(p, line, 256);
			n = tokenize(p, field, 16);
			if(n < 4)
				continue;
			if(match(path, field[0])){
				ok = 1;
				field[n] = nil;
				break;
			}
		}
		Bterm(b);
		if(!ok)
			return nil;
		h = halloc(c, sizeof(Handle));
		h->mimetype = field[1];
		h->ramfs = atoi(field[2]);
		/* we must have a room for additional $target at the end */
		v = halloc(c, n * sizeof(char*));
		for(n = 3; field[n]; n++)
			v[n - 3] = field[n];
		v[n -3] = nil;
		h->argv = v;
		return h;
	}
	return nil;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
static char *
hcat2(HConnect *c, char *a1, char *a2)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	return b;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
static char *
hcat3(HConnect *c, char *a1, char *a2, char *a3)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	strcat(b, a3);
	return b;
}

/*
*	we have two cases of uri
*	scheme://host/foo
*	/foo		# where should I go?
*
*	Note: in last case, hmoved() cares for putting scheme://host
*	but the original hmoved() can care only http.
*	must be fixed!
*	scheme should be kept in redirection.
*
*	It's better to close connection if destination is not
*	this name space.
*	uri:
*	"/~who/path"
*	"/path"
*/
static char *
addoptions(HConnect *c, char *uri)
{
	char *new;
	int n;
	char *p, *u;
	char *option;
	option = c->req.options;
	if(option == nil)
		option = "";
	u = nil; // user in uri
	if(uri[0] == '/'){
		if(uri[1] == '~'){
			p = strchr(uri+2, '/');
			if(p)
				*p = 0;
			u = hstrdup(c, uri+2);
			if(p)
				*p = '/';
		}
		if(xstrcmp(c->req.user, u) != 0)
			c->head.closeit = 1;
	}
	else
		c->head.closeit = 1;
	n = strlen(uri) + strlen(option) + 1 + UTFmax;
	new = halloc(c, n);
	snprint(new, n, "%s%s", uri, option);
	return new;
}

/*
*	printheaders() print HTTP headers.
*	status is exit-message of a CGI program
*	If stasus is nil or "" , it means normal exit.
*	CGI program should end with exit-message of some HTTP
*	response rule if it does not end normally.
*	However my old CGI programs are coded in loose manner;
*	they exits with groundless message.
*	Therefore printheader should handle these exit-message
*	so that it is compatible with older version of Pegasus:
*	if status is not HTTP response format we should assign
*		status = "200 OK"
*		c->head.closeit = 1
*/
static void
printheaders(HConnect *c, char *status, char *mimetype, int length)
{
	Hio *hout = &c->hout;
	char *p;
	int f=0; // f = 1 is status is HTTP response message
	if(status == nil || strlen(status) < 3)
		status = "200 OK";
	if(strlen(status) > 2
	&& isdigit(status[0]) &&  isdigit(status[1]) &&  isdigit(status[2]))
		f = 1;
	if(strlen(status) > 3 && !isspace(status[3]))
		f = 0;
	/*	we return "200 OK" for non HTTP exit message for compatibility	*/
	if(f == 0){
		status = "200 OK";
		c->head.closeit = 1;
	}
	else{
		status = hstrdup(c, status);
		p = strchr(status, '#');
		if(p){
			*p++ = 0;
			p = trim(p);
			if(cistrcmp(p, "close") == 0)
				c->head.closeit = 1;
			else if(cistrcmp(p, "keep") == 0)
				c->head.closeit = 0;
			status = trim(status);
		}
		else{
			if(status[0] == '1' || status[1] == '2')
				c->head.closeit = 0;
			else
				c->head.closeit = 1;
		}
	}

	hprint(hout, "%s %s\r\n", hversion, status);
	hprint(hout, "Server: %s\r\n", serversoft);
	hprint(hout, "Date: %D\r\n", time(nil));
	hprint(hout, "Content-Type: %s\r\n", mimetype);
	if(c->head.closeit)
		hprint(hout, "Connection: close\r\n");
	else if(!http11(c))
		hprint(hout, "Connection: Keep-Alive\r\n");
	/*	the rest will be sent later	*/
	hprint(hout, "Content-Length: %d\r\n", length);
	hprint(hout, "\r\n");
	hflush(hout);
}

/*	we are in emergency.
*	therefore we send a simple message to the client. sorry	*/
int
creject(int ctl, char *ndir)
{
	int data;
	int status = 0;
	char *msg ="\
HTTP/1.1 503 Service Unavailable\r\n\
Server: Plan9\r\n\
Content-Type: text/plain\r\n\
Content-Length: 5\r\n\
Connection: close\r\n\
\r\n\
Busy\n";

	switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
	case -1:
		status = -1;
		break;
	case 0:
		/*
		 *  see if we know the service requested
		*
		*	What is the first argument of accept should be ?
		*	man dial(2) says, it is nctl
		*	however /sys/src/cmd/ip/httpd/httpd.c says, ctl
		 */
		data = accept(ctl, ndir);
		if(data < 0){
			syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
			exits(nil);
		}
		write(data, msg, strlen(msg));
		close(data);
		exits(nil);
	default:
		break;
	}
	return status;
}
